"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DatabricksDriver = void 0;
/**
 * @copyright Cube Dev, Inc.
 * @license Apache-2.0
 * @fileoverview The `DatabricksDriver` and related types declaration.
 */
/* eslint-disable camelcase */
const shared_1 = require("@cubejs-backend/shared");
const client_s3_1 = require("@aws-sdk/client-s3");
const s3_request_presigner_1 = require("@aws-sdk/s3-request-presigner");
const storage_blob_1 = require("@azure/storage-blob");
const jdbc_driver_1 = require("@cubejs-backend/jdbc-driver");
const DatabricksQuery_1 = require("./DatabricksQuery");
const helpers_1 = require("./helpers");
const DatabricksToGenericType = {
    'decimal(10,0)': 'bigint',
};
/**
 * Databricks driver class.
 */
class DatabricksDriver extends jdbc_driver_1.JDBCDriver {
    static dialectClass() {
        return DatabricksQuery_1.DatabricksQuery;
    }
    /**
     * Returns default concurrency value.
     */
    static getDefaultConcurrency() {
        return 2;
    }
    /**
     * Class constructor.
     */
    constructor(conf = {}) {
        const dataSource = conf.dataSource ||
            (0, shared_1.assertDataSource)('default');
        let showSparkProtocolWarn = false;
        let url = conf?.url ||
            (0, shared_1.getEnv)('databrickUrl', { dataSource }) ||
            (0, shared_1.getEnv)('jdbcUrl', { dataSource });
        if (url.indexOf('jdbc:spark://') !== -1) {
            showSparkProtocolWarn = true;
            url = url.replace('jdbc:spark://', 'jdbc:databricks://');
        }
        const config = {
            ...conf,
            url,
            dbType: 'databricks',
            drivername: 'com.databricks.client.jdbc.Driver',
            customClassPath: undefined,
            properties: {
                UID: (0, helpers_1.extractUidFromJdbcUrl)(url),
                // PWD-parameter passed to the connection string has higher priority,
                // so we can set this one to an empty string to avoid a Java error.
                PWD: conf?.token ||
                    (0, shared_1.getEnv)('databrickToken', { dataSource }) ||
                    '',
                UserAgentEntry: 'CubeDev_Cube',
            },
            catalog: conf?.catalog ||
                (0, shared_1.getEnv)('databricksCatalog', { dataSource }),
            database: (0, shared_1.getEnv)('dbName', { required: false, dataSource }),
            // common export bucket config
            bucketType: conf?.bucketType ||
                (0, shared_1.getEnv)('dbExportBucketType', { supported: ['s3', 'azure'], dataSource }),
            exportBucket: conf?.exportBucket ||
                (0, shared_1.getEnv)('dbExportBucket', { dataSource }),
            exportBucketMountDir: conf?.exportBucketMountDir ||
                (0, shared_1.getEnv)('dbExportBucketMountDir', { dataSource }),
            pollInterval: (conf?.pollInterval ||
                (0, shared_1.getEnv)('dbPollMaxInterval', { dataSource })) * 1000,
            // AWS export bucket config
            awsKey: conf?.awsKey ||
                (0, shared_1.getEnv)('dbExportBucketAwsKey', { dataSource }),
            awsSecret: conf?.awsSecret ||
                (0, shared_1.getEnv)('dbExportBucketAwsSecret', { dataSource }),
            awsRegion: conf?.awsRegion ||
                (0, shared_1.getEnv)('dbExportBucketAwsRegion', { dataSource }),
            // Azure export bucket
            azureKey: conf?.azureKey ||
                (0, shared_1.getEnv)('dbExportBucketAzureKey', { dataSource }),
            exportBucketCsvEscapeSymbol: (0, shared_1.getEnv)('dbExportBucketCsvEscapeSymbol', { dataSource }),
        };
        super(config);
        this.config = config;
        this.showSparkProtocolWarn = showSparkProtocolWarn;
    }
    /**
     * @override
     */
    readOnly() {
        return !!this.config.readOnly;
    }
    /**
     * @override
     */
    capabilities() {
        return {
            unloadWithoutTempTable: true,
            incrementalSchemaLoading: true
        };
    }
    /**
     * @override
     */
    setLogger(logger) {
        super.setLogger(logger);
        this.showDeprecations();
    }
    /**
     * @override
     */
    async loadPreAggregationIntoTable(preAggregationTableName, loadSql, params, _options) {
        if (this.config.catalog) {
            const [schema] = preAggregationTableName.split('.');
            return super.loadPreAggregationIntoTable(preAggregationTableName, loadSql.replace(new RegExp(`(?<=\\s)${schema}\\.(?=[^\\s]+)`, 'g'), `${this.config.catalog}.${schema}.`), params, _options);
        }
        else {
            return super.loadPreAggregationIntoTable(preAggregationTableName, loadSql, params, _options);
        }
    }
    /**
     * @override
     */
    async query(query, values) {
        if (this.config.catalog) {
            return super.query(query.replace(new RegExp(`(?<=\\s)${this.getPreaggsSchemaName()}\\.(?=[^\\s]+)`, 'g'), `${this.config.catalog}.${this.getPreaggsSchemaName()}.`), values);
        }
        else {
            return super.query(query, values);
        }
    }
    /**
     * Returns pre-aggregation schema name.
     */
    getPreaggsSchemaName() {
        const schema = (0, shared_1.getEnv)('preAggregationsSchema');
        if (schema) {
            return schema;
        }
        else {
            const devMode = process.env.NODE_ENV !== 'production' || (0, shared_1.getEnv)('devMode');
            return devMode
                ? 'dev_pre_aggregations'
                : 'prod_pre_aggregations';
        }
    }
    /**
     * @override
     */
    dropTable(tableName, options) {
        const tableFullName = `${this.config?.catalog ? `${this.config.catalog}.` : ''}${tableName}`;
        return super.dropTable(tableFullName, options);
    }
    showDeprecations() {
        if (this.config.url) {
            const result = this.config.url
                .split(';')
                .find(node => /^PWD/i.test(node))
                ?.split('=')[1];
            if (result) {
                this.logger('PWD Parameter Deprecation in connection string', {
                    warning: 'PWD parameter is deprecated and will be ignored in future releases. ' +
                        'Please migrate to the CUBEJS_DB_DATABRICKS_TOKEN environment variable.'
                });
            }
        }
        if (this.showSparkProtocolWarn) {
            this.logger('jdbc:spark protocol deprecation', {
                warning: 'The `jdbc:spark` protocol is deprecated and will be ignored in future releases. ' +
                    'Please migrate your CUBEJS_DB_DATABRICKS_URL environment variable to the ' +
                    '`jdbc:databricks` protocol.'
            });
        }
    }
    /**
     * @override
     */
    async getCustomClassPath() {
        return (0, helpers_1.resolveJDBCDriver)();
    }
    /**
     * Execute create schema query.
     */
    async createSchemaIfNotExists(schemaName) {
        await this.query(`CREATE SCHEMA IF NOT EXISTS ${this.getSchemaFullName(schemaName)}`, []);
    }
    /**
     * Returns the list of the tables for the specified schema.
     */
    async getTablesQuery(schemaName) {
        const response = await this.query(`SHOW TABLES IN ${this.getSchemaFullName(schemaName)}`, []);
        return response.map((row) => ({
            table_name: row.tableName,
        }));
    }
    /**
     * Returns tables meta data object.
     */
    async tablesSchema() {
        const tables = await this.getTables();
        const metadata = {};
        await Promise.all(tables.map(async ({ database, tableName }) => {
            if (!(database in metadata)) {
                metadata[database] = {};
            }
            const columns = await this.tableColumnTypes(`${database}.${tableName}`);
            metadata[database][tableName] = columns;
        }));
        return metadata;
    }
    /**
     * Returns list of accessible tables.
     */
    async getTables() {
        if (this.config.database) {
            return this.query(`SHOW TABLES IN ${this.getSchemaFullName(this.config.database)}`, []);
        }
        const databases = await this.query(`SHOW DATABASES${this.config?.catalog
            ? ` IN ${this.quoteIdentifier(this.config.catalog)}`
            : ''}`, []);
        const tables = await Promise.all(databases.map(async ({ databaseName }) => this.query(`SHOW TABLES IN ${this.getSchemaFullName(databaseName)}`, [])));
        return tables.flat();
    }
    async getSchemas() {
        const databases = await this.query(`SHOW DATABASES${this.config?.catalog
            ? ` IN ${this.quoteIdentifier(this.config.catalog)}`
            : ''}`, []);
        return databases.map(({ databaseName }) => ({
            schema_name: databaseName,
        }));
    }
    async getTablesForSpecificSchemas(schemas) {
        const tables = await Promise.all(
        // eslint-disable-next-line camelcase
        schemas.map(async ({ schema_name }) => this.query(`SHOW TABLES IN ${this.getSchemaFullName(schema_name)}`, [])));
        return tables.flat().map(({ database, tableName }) => ({
            table_name: tableName,
            schema_name: database,
        }));
    }
    async getColumnsForSpecificTables(tables) {
        const columns = await Promise.all(
        // eslint-disable-next-line camelcase
        tables.map(async ({ schema_name, table_name }) => {
            const tableFullName = `${this.config?.catalog
                ? `${this.config.catalog}.`
                : ''
            // eslint-disable-next-line camelcase
            }${schema_name}.${table_name}`;
            const columnTypes = await this.tableColumnTypes(tableFullName);
            return columnTypes.map(({ name, type }) => ({
                column_name: name,
                data_type: type,
                table_name,
                schema_name,
            }));
        }));
        return columns.flat();
    }
    /**
     * Returns table columns types.
     */
    async tableColumnTypes(table) {
        let tableFullName = '';
        const tableArray = table.split('.');
        if (tableArray.length === 3) {
            tableFullName = `${this.quoteIdentifier(tableArray[0])}.${this.quoteIdentifier(tableArray[1])}.${this.quoteIdentifier(tableArray[2])}`;
        }
        else if (tableArray.length === 2 && this.config?.catalog) {
            tableFullName = `${this.quoteIdentifier(this.config.catalog)}.${this.quoteIdentifier(tableArray[0])}.${this.quoteIdentifier(tableArray[1])}`;
        }
        else {
            tableFullName = `${this.quoteIdentifier(tableArray[0])}.${this.quoteIdentifier(tableArray[1])}`;
        }
        const result = [];
        const response = await this.query(`DESCRIBE ${tableFullName}`, []);
        for (const column of response) {
            // Databricks describe additional info by default after empty line.
            if (column.col_name === '') {
                break;
            }
            result.push({
                name: column.col_name,
                type: this.toGenericType(column.data_type),
            });
        }
        return result;
    }
    /**
     * Returns query columns types.
     */
    async queryColumnTypes(sql, params) {
        const result = [];
        // eslint-disable-next-line camelcase
        const response = await this.query(`DESCRIBE QUERY ${sql}`, params || []);
        for (const column of response) {
            // Databricks describe additional info by default after empty line.
            if (column.col_name === '') {
                break;
            }
            result.push({ name: column.col_name, type: this.toGenericType(column.data_type) });
        }
        return result;
    }
    /**
     * Returns schema full name.
     */
    getSchemaFullName(schema) {
        if (this.config?.catalog) {
            return `${this.quoteIdentifier(this.config.catalog)}.${this.quoteIdentifier(schema)}`;
        }
        else {
            return `${this.quoteIdentifier(schema)}`;
        }
    }
    /**
     * Returns quoted string.
     */
    quoteIdentifier(identifier) {
        return `\`${identifier}\``;
    }
    /**
     * Returns the JS type by the Databricks type.
     */
    toGenericType(columnType) {
        return DatabricksToGenericType[columnType.toLowerCase()] || super.toGenericType(columnType);
    }
    /**
     * Determines whether export bucket feature is configured or not.
     */
    async isUnloadSupported() {
        return this.config.exportBucket !== undefined;
    }
    /**
     * Returns to the Cubestore an object with links to unloaded to an
     * export bucket data.
     */
    async unload(tableName, options) {
        if (!['azure', 's3'].includes(this.config.bucketType)) {
            throw new Error(`Unsupported export bucket type: ${this.config.bucketType}`);
        }
        const tableFullName = `${this.config.catalog
            ? `${this.config.catalog}.`
            : ''}${tableName}`;
        const types = options.query
            ? await this.unloadWithSql(tableFullName, options.query.sql, options.query.params)
            : await this.unloadWithTable(tableFullName);
        const pathname = `${this.config.exportBucket}/${tableFullName}.csv`;
        const csvFile = await this.getCsvFiles(pathname);
        return {
            exportBucketCsvEscapeSymbol: this.config.exportBucketCsvEscapeSymbol,
            csvFile,
            types,
            csvNoHeader: true,
        };
    }
    /**
     * Unload data from a SQL query to an export bucket.
     */
    async unloadWithSql(tableFullName, sql, params) {
        const types = await this.queryColumnTypes(sql, params);
        await this.createExternalTableFromSql(tableFullName, sql, params);
        return types;
    }
    /**
     * Unload data from a temp table to an export bucket.
     */
    async unloadWithTable(tableFullName) {
        const types = await this.tableColumnTypes(tableFullName);
        const columns = types.map(t => t.name).join(', ');
        await this.createExternalTableFromTable(tableFullName, columns);
        return types;
    }
    /**
     * Returns an array of signed URLs of unloaded csv files.
     */
    async getCsvFiles(pathname) {
        let res;
        switch (this.config.bucketType) {
            case 'azure':
                res = await this.getSignedAzureUrls(pathname);
                break;
            case 's3':
                res = await this.getSignedS3Urls(pathname);
                break;
            default:
                throw new Error(`Unsupported export bucket type: ${this.config.bucketType}`);
        }
        return res;
    }
    /**
     * Returns Azure signed URLs of unloaded scv files.
     */
    async getSignedAzureUrls(pathname) {
        const csvFile = [];
        const [container, account] = pathname.split('wasbs://')[1].split('.blob')[0].split('@');
        const foldername = pathname.split(`${this.config.exportBucket}/`)[1];
        const expr = new RegExp(`${foldername}\\/.*\\.csv$`, 'i');
        const credential = new storage_blob_1.StorageSharedKeyCredential(account, this.config.azureKey);
        const blobClient = new storage_blob_1.BlobServiceClient(`https://${account}.blob.core.windows.net`, credential);
        const containerClient = blobClient.getContainerClient(container);
        const blobsList = containerClient.listBlobsFlat({ prefix: foldername });
        for await (const blob of blobsList) {
            if (blob.name && expr.test(blob.name)) {
                const sas = (0, storage_blob_1.generateBlobSASQueryParameters)({
                    containerName: container,
                    blobName: blob.name,
                    permissions: storage_blob_1.ContainerSASPermissions.parse('r'),
                    startsOn: new Date(new Date().valueOf()),
                    expiresOn: new Date(new Date().valueOf() + 1000 * 60 * 60),
                    protocol: storage_blob_1.SASProtocol.Https,
                    version: '2020-08-04',
                }, credential).toString();
                csvFile.push(`https://${account}.blob.core.windows.net/${container}/${blob.name}?${sas}`);
            }
        }
        if (csvFile.length === 0) {
            throw new Error('No CSV files were exported to the specified bucket. ' +
                'Please check your export bucket configuration.');
        }
        return csvFile;
    }
    /**
     * Returns S3 signed URLs of unloaded scv files.
     */
    async getSignedS3Urls(pathname) {
        const client = new client_s3_1.S3({
            credentials: {
                accessKeyId: this.config.awsKey,
                secretAccessKey: this.config.awsSecret,
            },
            region: this.config.awsRegion,
        });
        const url = new URL(pathname);
        const list = await client.listObjectsV2({
            Bucket: url.host,
            Prefix: url.pathname.slice(1),
        });
        if (list.Contents === undefined) {
            throw new Error(`No content in specified path: ${pathname}`);
        }
        const csvFile = await Promise.all(list.Contents
            .filter(file => file.Key && /.csv$/i.test(file.Key))
            .map(async (file) => {
            const command = new client_s3_1.GetObjectCommand({
                Bucket: url.host,
                Key: file.Key,
            });
            return (0, s3_request_presigner_1.getSignedUrl)(client, command, { expiresIn: 3600 });
        }));
        if (csvFile.length === 0) {
            throw new Error('No CSV files were exported to the specified bucket. ' +
                'Please check your export bucket configuration.');
        }
        return csvFile;
    }
    /**
     * Saves specified query to the configured bucket. This requires Databricks
     * cluster to be configured.
     *
     * For Azure blob storage you need to configure account access key in
     * Cluster -> Configuration -> Advanced options
     * (https://docs.databricks.com/data/data-sources/azure/azure-storage.html#access-azure-blob-storage-directly)
     *
     * `fs.azure.account.key.<storage-account-name>.blob.core.windows.net <storage-account-access-key>`
     *
     * For S3 bucket storage you need to configure AWS access key and secret in
     * Cluster -> Configuration -> Advanced options
     * (https://docs.databricks.com/data/data-sources/aws/amazon-s3.html#access-s3-buckets-directly)
     *
     * `fs.s3a.access.key <aws-access-key>`
     * `fs.s3a.secret.key <aws-secret-key>`
     */
    async createExternalTableFromSql(tableFullName, sql, params) {
        try {
            await this.query(`
        CREATE TABLE ${tableFullName}
        USING CSV LOCATION '${this.config.exportBucketMountDir || this.config.exportBucket}/${tableFullName}.csv'
        OPTIONS (escape = '"')
        AS (${sql});
        `, params);
        }
        finally {
            await this.query(`DROP TABLE IF EXISTS ${tableFullName};`, []);
        }
    }
    /**
     * Saves specified table to the configured bucket. This requires Databricks
     * cluster to be configured.
     *
     * For Azure blob storage you need to configure account access key in
     * Cluster -> Configuration -> Advanced options
     * (https://docs.databricks.com/data/data-sources/azure/azure-storage.html#access-azure-blob-storage-directly)
     *
     * `fs.azure.account.key.<storage-account-name>.blob.core.windows.net <storage-account-access-key>`
     *
     * For S3 bucket storage you need to configure AWS access key and secret in
     * Cluster -> Configuration -> Advanced options
     * (https://docs.databricks.com/data/data-sources/aws/amazon-s3.html#access-s3-buckets-directly)
     *
     * `fs.s3a.access.key <aws-access-key>`
     * `fs.s3a.secret.key <aws-secret-key>`
     */
    async createExternalTableFromTable(tableFullName, columns) {
        try {
            await this.query(`
        CREATE TABLE _${tableFullName}
        USING CSV LOCATION '${this.config.exportBucketMountDir || this.config.exportBucket}/${tableFullName}.csv'
        OPTIONS (escape = '"')
        AS SELECT ${columns} FROM ${tableFullName}
        `, []);
        }
        finally {
            await this.query(`DROP TABLE IF EXISTS _${tableFullName};`, []);
        }
    }
}
exports.DatabricksDriver = DatabricksDriver;
//# sourceMappingURL=DatabricksDriver.js.map