"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.DatabricksQuery = void 0;
const ramda_1 = __importDefault(require("ramda"));
const schema_compiler_1 = require("@cubejs-backend/schema-compiler");
const GRANULARITY_TO_INTERVAL = {
    day: 'day',
    week: 'week',
    hour: 'hour',
    minute: 'minute',
    second: 'second',
    month: 'month',
    quarter: 'quarter',
    year: 'year'
};
class DatabricksFilter extends schema_compiler_1.BaseFilter {
    likeIgnoreCase(column, not, param, type) {
        const p = (!type || type === 'contains' || type === 'ends') ? '%' : '';
        const s = (!type || type === 'contains' || type === 'starts') ? '%' : '';
        return `LOWER(${column})${not ? ' NOT' : ''} LIKE CONCAT('${p}', LOWER(${this.allocateParam(param)}), '${s}')`;
    }
}
class DatabricksQuery extends schema_compiler_1.BaseQuery {
    newFilter(filter) {
        return new DatabricksFilter(this, filter);
    }
    convertTz(field) {
        return `from_utc_timestamp(${field}, '${this.timezone}')`;
    }
    timeStampCast(value) {
        return `from_utc_timestamp(replace(replace(${value}, 'T', ' '), 'Z', ''), 'UTC')`;
    }
    dateTimeCast(value) {
        return `from_utc_timestamp(${value}, 'UTC')`; // TODO
    }
    subtractInterval(date, interval) {
        const [number, type] = this.parseInterval(interval);
        return `(${date} - INTERVAL '${number}' ${type})`;
    }
    addInterval(date, interval) {
        const [number, type] = this.parseInterval(interval);
        return `(${date} + INTERVAL '${number}' ${type})`;
    }
    timeGroupedColumn(granularity, dimension) {
        return `date_trunc('${GRANULARITY_TO_INTERVAL[granularity]}', ${dimension})`;
    }
    escapeColumnName(name) {
        return `\`${name}\``;
    }
    getFieldIndex(id) {
        const dimension = this.dimensionsForSelect().find((d) => d.dimension === id);
        if (dimension) {
            return super.getFieldIndex(id);
        }
        return this.escapeColumnName(this.aliasName(id, false));
    }
    unixTimestampSql() {
        return 'unix_timestamp()';
    }
    orderHashToString(hash) {
        if (!hash || !hash.id) {
            return null;
        }
        const fieldIndex = this.getFieldIndex(hash.id);
        if (fieldIndex === null) {
            return null;
        }
        const dimensionsForSelect = this.dimensionsForSelect();
        const dimensionColumns = ramda_1.default.flatten(dimensionsForSelect.map((s) => s.selectColumns() && s.aliasName()))
            .filter(s => !!s);
        if (dimensionColumns.length) {
            const direction = hash.desc ? 'DESC' : 'ASC';
            return `${fieldIndex} ${direction}`;
        }
        return null;
    }
    defaultRefreshKeyRenewalThreshold() {
        return 120;
    }
    sqlTemplates() {
        const templates = super.sqlTemplates();
        templates.functions.DATETRUNC = 'DATE_TRUNC({{ args_concat }})';
        templates.functions.BTRIM = 'TRIM({% if args[1] is defined %}{{ args[1] }} FROM {% endif %}{{ args[0] }})';
        templates.functions.LTRIM = 'LTRIM({{ args|reverse|join(", ") }})';
        templates.functions.RTRIM = 'RTRIM({{ args|reverse|join(", ") }})';
        templates.functions.DATEDIFF = 'DATEDIFF({{ date_part }}, DATE_TRUNC(\'{{ date_part }}\', {{ args[1] }}), DATE_TRUNC(\'{{ date_part }}\', {{ args[2] }}))';
        templates.expressions.timestamp_literal = 'from_utc_timestamp(\'{{ value }}\', \'UTC\')';
        return templates;
    }
}
exports.DatabricksQuery = DatabricksQuery;
//# sourceMappingURL=DatabricksQuery.js.map